/***********************************************************************
*   Author: Philip Thompson (phils@athena.mit.edu)
*   - convert from RADIANCE image file format to Xim and vice-versa.
*  Copyright (c) 1990, Massachusetts Institute of Technology
*                Philip R. Thompson (phils@athena.mit.edu)
*                Computer Resource Laboratory (CRL)
*                Dept. of Architecture and Planning
*                M.I.T., Rm 9-526
*                Cambridge, MA  02139
*   This  software and its documentation may be used, copied, modified,
*   and distributed for any purpose without fee, provided:
*       --  The above copyright notice appears in all copies.
*       --  This disclaimer appears in all source code copies.
*       --  The names of M.I.T. and the CRL are not used in advertising
*           or publicity pertaining to distribution of the software
*           without prior specific written permission from me or CRL.
*   I provide this software freely as a public service.  It is NOT
*   public domain nor a commercial product, and therefore is not subject
*   to an an implied warranty of merchantability or fitness for a
*   particular purpose.  I provide it as is, without warranty.
*   This software is furnished  only on the basis that any party who
*   receives it indemnifies and holds harmless the parties who furnish
*   it against any claims, demands, or liabilities connected with using
*   it, furnishing it to adatas, or providing it to a third party.
***********************************************************************/
#if (!defined(lint) && !defined(SABER))
static char radiance_rcsid[] =
"$Header: /mit/crldev/src/xim/lib/RCS/radiance_io.c,v 1.16 92/06/22 11:03:38 phils Exp Locker: phils $";
#endif

/* Copyright (c) 1991 Regents of the University of California */
/*
 *  color.c - routines for color calculations.
 *     10/10/85
 *  header.c - routines for reading and writing information headers.
 *      8/19/88
 */

#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <pwd.h>
#include "xim.h"
#include "ximdefs.h"
#include "ximproto.h"
#include <sys/time.h>

#define  RED            0
#define  GRN            1
#define  BLU            2
#define  EXP            3
#define  COLXS          128     /* excess used for exponent */

#define XDECR   1
#define YDECR   2
#define YMAJOR  4
#define MINELEN  8            /* minimum scanline length for encoding */
#define MINRUN   4            /* minimum run length */
#define MAXGSHIFT   15        /* maximum shift for gamma table */
#define MAXLINE     512
#define FMTSTRL 7
#define FMTSTR  "FORMAT="
#define COLRFMT "32-bit_rle_rgbe"
#define copycolr(c1,c2)     (c1[0]=c2[0],c1[1]=c2[1],c1[2]=c2[2],\
                            c1[3]=c2[3])

typedef byte COLR[4];           /* red, green, blue, exponent */
typedef struct {
    int or;         /* orientation (from flags above) */
    int xr, yr;     /* x and y resolution */
} RESOLU;
struct check {
    FILE *fp;
    char fs[64];
};

extern int ximDebugFlag;


/* return format value */
static void formatval(r, s)
register char *r, *s;
{
    s += FMTSTRL;
    while (isspace(*s))
        s++;
    if (!*s) {
        *r = '\0';
        return;
    }
    while (*s)
        *r++ = *s++;
    while (isspace(r[-1]))
        r--;
    *r = '\0';
}


/* put out a format value */
static void fputformat(s, fp)
char *s;
FILE *fp;
{
    fputs(FMTSTR, fp);
    fputs(s, fp);
    putc('\n', fp);
}


/* get header from file */
static int getheader(fp, f, p)
FILE *fp;
void (*f)();
struct check *p;
{
    char buf[MAXLINE];

    for (;;) {
        buf[MAXLINE - 2] = '\n';
        if (fgets(buf, sizeof(buf), fp) == NULL)
            return (-1);
        if (buf[0] == '\n')
            return 0;
        if (buf[MAXLINE - 2] != '\n') {
            ungetc(buf[MAXLINE - 2], fp);       /* prevent false end */
            buf[MAXLINE - 2] = '\0';
        }
        if (f != NULL)
            (*f)(buf, p);
    }
    return 1;
}


/* check a header line for format info. */
static void mycheck(s, cp)
char *s;
struct check *cp;
{
    if (!strncmp(s, FMTSTR, FMTSTRL))
        formatval(cp->fs, s);
    else if (cp->fp != NULL)    /* don't copy format info. */
        fputs(s, cp->fp);
}

#define copymatch(pat, s)       (!strcmp(pat, s))

/* Checkheader(fin,fmt,fout) returns a value of 1 if the input format
 * matches the specification in fmt, 0 if no input format was found,
 * and -1 if the input format does not match or there is an
 * error reading the header.  If fmt is empty, then -1 is returned
 * if any input format is found (or there is an error), and 0 otherwise.
 * If fmt contains any '*' or '?' characters, then checkheader
 * does wildcard expansion and copies a matching result into fmt.
 * Be sure that fmt is big enough to hold the match in such cases!
 * The input header (minus any format lines) is copied to fout
 * if fout is not NULL.
 */
static int checkheader(fin, fmt, fout)
FILE *fin;
char *fmt;
FILE *fout;
{
    struct check cdat;

    cdat.fp = fout;
    cdat.fs[0] = '\0';
    if (getheader(fin, mycheck, &cdat) < 0)
        return -1;
    if (cdat.fs[0] != '\0')
        return (copymatch(fmt, cdat.fs) ? 1 : -1);
    return 0;
}


/* set gamma conversion */
static void setcolrgam(g, g_mant, g_nexp, g_bval)
double g;
byte g_mant[256], g_nexp[256], g_bval[MAXGSHIFT+1][256];
{
    double mult;
    register int i, j;

    /* compute colr -> gamb mapping */
    for (i = 0; i <= MAXGSHIFT; i++) {
        mult = pow(0.5, (double)(i + 8));
        for (j = 0; j < 256; j++)
            g_bval[i][j] = 256.0 * pow((j + .5) * mult, 1.0 / g);
    }
    /* compute gamb -> colr mapping */
    i = 0;
    mult = 256.0;
    for (j = 255; j > 0; j--) {
        while ((g_mant[j] = mult * pow(j / 256.0, g)) < 128) {
            i++;
            mult *= 2.0;
        }
        g_nexp[j] = i;
    }
    g_mant[0] = 0;
    g_nexp[0] = COLXS;
}


/* convert scanline of colrs to gamma bytes */
static void colrs_gambs(scan, len, g_mant, g_nexp, g_bval)
register COLR *scan;
int len;
byte g_mant[256], g_nexp[256], g_bval[MAXGSHIFT+1][256];
{
    register int i, expo;

    while (len-- > 0) {
        expo = scan[0][EXP] - COLXS;
        if (expo < -MAXGSHIFT) {
            if (expo < -MAXGSHIFT - 8) {
                scan[0][RED] = scan[0][GRN] = scan[0][BLU] = 0;
            } else {
                i = (-MAXGSHIFT - 1) - expo;
                scan[0][RED] =
                        g_bval[MAXGSHIFT][((scan[0][RED] >> i) + 1) >> 1];
                scan[0][GRN] =
                        g_bval[MAXGSHIFT][((scan[0][GRN] >> i) + 1) >> 1];
                scan[0][BLU] =
                        g_bval[MAXGSHIFT][((scan[0][BLU] >> i) + 1) >> 1];
            }
        } else if (expo > 0) {
            if (expo > 8) {
                scan[0][RED] = scan[0][GRN] = scan[0][BLU] = 255;
            } else {
                i = (scan[0][RED] << 1 | 1) << (expo - 1);
                scan[0][RED] = i > 255 ? 255 : g_bval[0][i];
                i = (scan[0][GRN] << 1 | 1) << (expo - 1);
                scan[0][GRN] = i > 255 ? 255 : g_bval[0][i];
                i = (scan[0][BLU] << 1 | 1) << (expo - 1);
                scan[0][BLU] = i > 255 ? 255 : g_bval[0][i];
            }
        } else {
            scan[0][RED] = g_bval[-expo][scan[0][RED]];
            scan[0][GRN] = g_bval[-expo][scan[0][GRN]];
            scan[0][BLU] = g_bval[-expo][scan[0][BLU]];
        }
        scan[0][EXP] = COLXS;
        scan++;
    }
}


/* convert gamma bytes to col scanline */
static void gambs_colrs(scan, len, g_mant, g_nexp, g_bval)
register COLR *scan;
int len;
byte g_mant[256], g_nexp[256], g_bval[MAXGSHIFT+1][256];
{
    register int nexpo;

    while (len-- > 0) {
        nexpo = g_nexp[scan[0][RED]];
        if (g_nexp[scan[0][GRN]] < nexpo)
            nexpo = g_nexp[scan[0][GRN]];
        if (g_nexp[scan[0][BLU]] < nexpo)
            nexpo = g_nexp[scan[0][BLU]];
        if (nexpo < g_nexp[scan[0][RED]])
            scan[0][RED] = g_mant[scan[0][RED]] >>
					(g_nexp[scan[0][RED]] - nexpo);
        else
            scan[0][RED] = g_mant[scan[0][RED]];
        if (nexpo < g_nexp[scan[0][GRN]])
            scan[0][GRN] = g_mant[scan[0][GRN]] >>
					(g_nexp[scan[0][GRN]] - nexpo);
        else
            scan[0][GRN] = g_mant[scan[0][GRN]];
        if (nexpo < g_nexp[scan[0][BLU]])
            scan[0][BLU] = g_mant[scan[0][BLU]] >>
					(g_nexp[scan[0][BLU]] - nexpo);
        else
            scan[0][BLU] = g_mant[scan[0][BLU]];
        scan[0][EXP] = COLXS - nexpo;
        scan++;
    }
}

/* shift a scanline of colors by 2^adjust */
static void shiftcolrs(scan, len, adjust)
register COLR *scan;
register int len;
register int adjust;
{
    int minexp;

    if (adjust == 0)
        return;
    minexp = adjust < 0 ? -adjust : 0;
    while (len-- > 0) {
        if (scan[0][EXP] <= minexp)
            scan[0][RED] = scan[0][GRN] = scan[0][BLU] =
                    scan[0][EXP] = 0;
        else
            scan[0][EXP] += adjust;
        scan++;
    }
}


/* read in an old colr scanline */
static int oldreadcolrs(scanline, len, fp)
register COLR *scanline;
int len;
register FILE *fp;
{
    int rshift = 0;
    register int i;

    while (len > 0) {
        scanline[0][RED] = getc(fp);
        scanline[0][GRN] = getc(fp);
        scanline[0][BLU] = getc(fp);
        scanline[0][EXP] = getc(fp);
        if (feof(fp) || ferror(fp))
            return (-1);
        if (scanline[0][RED] == 1 &&
                scanline[0][GRN] == 1 &&
                scanline[0][BLU] == 1) {
            for (i = scanline[0][EXP] << rshift; i > 0; i--) {
                copycolr(scanline[0], scanline[-1]);
                scanline++;
                len--;
            }
            rshift += 8;
        } else {
            scanline++;
            len--;
            rshift = 0;
        }
    }
    return 0;
}


/* write out a colr scanline */
static int fwritecolrs(scanline, len, fp)
register COLR *scanline;
int len;
register FILE *fp;
{
    register int i, j, beg, cnt = 0;
    int c2;

    if (len < MINELEN)          /* too small to encode */
        return (fwrite((char*)scanline, sizeof(COLR), len, fp) - len);
    if (len > 32767)            /* too big! */
        return (-1);
    putc(2, fp);                /* put magic header */
    putc(2, fp);
    putc(len >> 8, fp);
    putc(len & 255, fp);
    /* put components seperately */
    for (i = 0; i < 4; i++) {
        for (j = 0; j < len; j += cnt) {        /* find next run */
            for (beg = j; beg < len; beg += cnt) {
                for (cnt = 1; cnt < 127 && beg + cnt < len &&
                        scanline[beg + cnt][i] == scanline[beg][i]; cnt++) ;
                if (cnt >= MINRUN)
                    break;      /* long enough */
            }
            if (beg - j > 1 && beg - j < MINRUN) {
                c2 = j + 1;
                while (scanline[c2++][i] == scanline[j][i])
                    if (c2 == beg) {    /* short run */
                        putc(128 + beg - j, fp);
                        putc(scanline[j][i], fp);
                        j = beg;
                        break;
                    }
            }
            while (j < beg) {   /* write out non-run */
                if ((c2 = beg - j) > 128)
                    c2 = 128;
                putc(c2, fp);
                while (c2--)
                    putc(scanline[j++][i], fp);
            }
            if (cnt >= MINRUN) {/* write out run */
                putc(128 + cnt, fp);
                putc(scanline[beg][i], fp);
            } else
                cnt = 0;
        }
    }
    return (ferror(fp) ? -1 : 0);
}


/* read in an encoded colr scanline */
static int freadcolrs(scanline, len, fp)
register COLR *scanline;
int len;
register FILE *fp;
{
    register int i, j;
    int code;

    /* determine scanline type */
    if (len < MINELEN)
        return (oldreadcolrs(scanline, len, fp));
    if ((i = getc(fp)) == EOF)
        return (-1);
    if (i != 2) {
        ungetc(i, fp);
        return (oldreadcolrs(scanline, len, fp));
    }
    scanline[0][GRN] = getc(fp);
    scanline[0][BLU] = getc(fp);
    if ((i = getc(fp)) == EOF)
        return (-1);
    if (scanline[0][GRN] != 2 || scanline[0][BLU] & 128) {
        scanline[0][RED] = 2;
        scanline[0][EXP] = i;
        return (oldreadcolrs(scanline + 1, len - 1, fp));
    }
    if ((scanline[0][BLU] << 8 | i) != len)
        return (-1);            /* length mismatch! */
    /* read each component */
    for (i = 0; i < 4; i++)
        for (j = 0; j < len;) {
            if ((code = getc(fp)) == EOF)
                return (-1);
            if (code > 128) {   /* run */
                scanline[j++][i] = getc(fp);
                for (code &= 127; --code; j++)
                    scanline[j][i] = scanline[j - 1][i];
            } else              /* non-run */
                while (code--)
                    scanline[j++][i] = getc(fp);
        }
    return (feof(fp) ? -1 : 0);
}


/* convert resolution line to struct */
static int str2resolu(rp, buf)
register RESOLU *rp;
char *buf;
{
    register char *xndx, *yndx, *cp;

    if (buf == NULL)
        return 0;
    xndx = yndx = NULL;
    for (cp = buf; *cp; cp++)
        if (*cp == 'X')
            xndx = cp;
        else if (*cp == 'Y')
            yndx = cp;
    if ((xndx == NULL) || (yndx == NULL))
        return 0;
    rp->or = 0;
    if (xndx > yndx)
        rp->or |= YMAJOR;
    if (xndx[-1] == '-')
        rp->or |= XDECR;
    if (yndx[-1] == '-')
        rp->or |= YDECR;
    if ((rp->xr = atoi(xndx + 1)) <= 0)
        return 0;
    if ((rp->yr = atoi(yndx + 1)) <= 0)
        return 0;
    return 1;
}

#define RESOLU_BUFLEN  32

/* get picture dimensions */
static int fgetresolu(sl, ns, fp)
int *sl, *ns;                   /* scanline length and number */
FILE *fp;
{
    RESOLU rs;
    char resolu_buf[RESOLU_BUFLEN];

    if (!str2resolu(&rs, fgets(resolu_buf, RESOLU_BUFLEN, fp)))
        return -1;
    if (rs.or & YMAJOR) {
        *sl = rs.xr;
        *ns = rs.yr;
    } else {
        *sl = rs.yr;
        *ns = rs.xr;
    }
    return rs.or;
}


static int RadianceWriteHeader(fp, x, y, xim)
FILE *fp;
int x, y;
Xim *xim;
{
    fprintf(fp, "Xim format conversion by: %s\n", xim->author);
    fputformat(COLRFMT, fp);
    fputc('\n', fp);
    fprintf(fp, "-Y %d +X %d\n", y, x);
    fflush(fp);
    return 1;
}


int XimReadRadiance(ifp, xim, gamma, bradj)
FILE *ifp;
Xim *xim;
double gamma;
int bradj;                      /* brightness adjust */
{
    register int x, y, i;
    int width, height;
    COLR *scanline;             /* RADIANCE variable */
    struct timeval time_val;
    struct timezone time_zone;
    struct passwd *pwd = NULL;
    char *login_ptr = NULL, *ctime(), *getlogin();
    byte g_mant[256], g_nexp[256], g_bval[MAXGSHIFT+1][256];

    /* get width, height from radiance file */
    if ((checkheader(ifp, COLRFMT, NULL) < 0) ||
            (fgetresolu(&width, &height, ifp) < 0)) {
        XimWarn("XimReadRadiance: bad file header");
        return 0;
    }
    if ((scanline = (COLR*)calloc(width, sizeof(COLR))) == NULL) {
        XimWarn("XimReadRadiance: can't calloc scanline");
        return 0;
    }
    setcolrgam(gamma, g_mant, g_nexp, g_bval);
    BZERO(xim, sizeof(Xim));
    xim->width = xim->bytes_line = width;
    xim->height = height;
    xim->nchannels = 3;
    xim->bits_channel = 8;
    xim->datasize = xim->height * xim->bytes_line;
    xim->npics = 1;
    xim->data = (byte*)malloc(xim->datasize);
    xim->gdata = (byte*)malloc(xim->datasize);
    xim->bdata = (byte*)malloc(xim->datasize);
    if (!xim->data || !xim->gdata || !xim->bdata) {
        FREE(xim->data);
        FREE(xim->gdata);
        FREE(xim->bdata);
    	xim->datasize = 0;
        XimWarn("XimReadRadiance: can't malloc xim buffers");
        return 0;
    }
    for (y = 0, i = 0; y < height; y++) {
        if (freadcolrs(scanline, width, ifp) < 0) {
            XimWarn("XimReadRadiance: read error");
            free((char*)scanline);
            break;
        }
        if (bradj)
            shiftcolrs(scanline, width, bradj);
        colrs_gambs(scanline, width, g_mant, g_nexp, g_bval);
        for (x = 0; x < width; x++, i++) {
            xim->data[i] = (byte)(scanline[x][RED]);
            xim->gdata[i] = (byte)(scanline[x][GRN]);
            xim->bdata[i] = (byte)(scanline[x][BLU]);
        }
        if (ferror(ifp)) {
            XimWarn("XimReadRadiance: read error");
            free((char*)scanline);
            break;
        }
    }
    free((char*)scanline);
	if ((y > 0) && (y < height)) { /* less than full image was read */
    	xim->height = height = y;
    	xim->datasize = xim->height * xim->bytes_line;
    	xim->data = (byte*)realloc(xim->data, xim->datasize);
    	xim->gdata = (byte*)realloc(xim->gdata, xim->datasize);
    	xim->bdata = (byte*)realloc(xim->bdata, xim->datasize);
    	if (!xim->data || !xim->gdata || !xim->bdata) {
       		FREE(xim->data);
       		FREE(xim->gdata);
       		FREE(xim->bdata);
    		xim->datasize = 0;
       		XimWarn("XimReadRadiance: can't realloc xim buffers");
       		return 0;
    	}
	}
    (void) gettimeofday(&time_val, &time_zone);
    xim->date = ctime(&time_val.tv_sec);
    xim->date[strlen(xim->date) - 1] = '\0';
    /* date points to static memory, so copy it to make freeable */
    xim->date = XimAllocAndStrCpy(xim->date);
    if ((login_ptr = getlogin()) != NULL)
        pwd = getpwnam(login_ptr);
    else
        pwd = getpwuid(getuid());
    if (pwd && !(xim->author = XimAllocAndStrCpy(pwd->pw_gecos))) {
        XimWarn("XimReadRadiance: can't alloc xim->author");
        return 0;
    } else
        xim->author = NULL;
    return 1;
}


int XimWriteRadiance(ofp, xim, gamma)
FILE *ofp;
Xim *xim;
double gamma;
{
    COLR *scanline;
    int width, height;
    register int x, y;
    register byte *red, *grn, *blu;
    byte g_mant[256], g_nexp[256], g_bval[MAXGSHIFT+1][256];

	if (!XIM_IS_24BIT(xim)) {
		XimWarn("XimWriteRadiance: xim not 24-bit");
		return 0;
	}
    /* get width, height from xim */
    width = xim->width;
    height = xim->height;

    /* allocate scan line space */
    if ((scanline = (COLR*)calloc(width, sizeof(COLR))) == NULL) {
        XimWarn("XimWriteRadiance: cant' alloc scanline");
        return 0;
    }
    setcolrgam(gamma, g_mant, g_nexp, g_bval);
    /* write out the RADIANCE header */
    if (!RadianceWriteHeader(ofp, width, height, xim)) {
        XimWarn("XimWriteRadiance: cant' alloc scanline");
        free((char*)scanline);
        return 0;
    }
    red = xim->data;
    grn = xim->gdata;
    blu = xim->bdata;
    for (y = 0; y < height; y++) {
        for (x = 0; x < width; x++) {
            scanline[x][RED] = red[x];
            scanline[x][GRN] = grn[x];
            scanline[x][BLU] = blu[x];
        }
        gambs_colrs(scanline, width, g_mant, g_nexp, g_bval);
        if (fwritecolrs(scanline, width, ofp) < 0) {
            XimWarn("XimWriteRadiance: write error");
            free((char*)scanline);
            return 0;
        }
        red += xim->bytes_line;
        grn += xim->bytes_line;
        blu += xim->bytes_line;
    }
    free((char*)scanline);
    return 1;
}

/*** end radiance_io.c ***/
